(* Author: Christina Kohl and René Thiemann *)

(* this theory provides executable versions for various rewrite relations, including
  parallel rewriting, multistep rewriting and parallel rewriting with variable restrictions *)
theory Rewrite_Relations_Impl
  imports 
    TRS.Trs_Impl
    Par_Step_Var_Restricted
    TRS.Multistep
    Proof_Terms.Utils
begin 


subsection \<open>Implementation of parallel rewriting with variable restriction\<close>

context 
  fixes R :: "('f,'v)rules" and V :: "'v set"  
begin
fun is_par_rstep_var_restr :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
where
  "is_par_rstep_var_restr (Fun f ss) (Fun g ts) =
    (Fun f ss = Fun g ts \<or> 
    vars_term (Fun g ts) \<inter> V = {} \<and> (Fun f ss, Fun g ts) \<in> rrstep (set R) \<or>
    (f = g \<and> length ss = length ts \<and> list_all2 is_par_rstep_var_restr ss ts))"
| "is_par_rstep_var_restr s t = (s = t \<or> vars_term t \<inter> V = {} \<and> (s,t) \<in> rrstep (set R))"

lemma is_par_rstep_code_helper: "vars_term t \<inter> V = {} \<longleftrightarrow>
  (\<forall> x \<in> set (vars_term_list t). x \<notin> V)" 
  by auto

lemmas is_par_rstep_var_restr_code[code] = is_par_rstep_var_restr.simps[unfolded is_par_rstep_code_helper]

lemma is_par_rstep_var_restr[simp]:
  "is_par_rstep_var_restr s t \<longleftrightarrow> (s, t) \<in> par_rstep_var_restr (set R) V"
proof 
  let ?Prop = "\<lambda> s t. s = t \<or> vars_term t \<inter> V = {} \<and> (s,t) \<in> rrstep (set R)" 
  {
    fix s t
    assume "?Prop s t" 
    hence "\<exists> C infos. (s, t) \<in> par_rstep_mctxt (set R) C infos \<and> vars_below_hole t C \<inter> V = {}"
    proof 
      assume "s = t" 
      thus ?thesis by (intro exI[of _ "mctxt_of_term s"] exI[of _ Nil], auto simp: par_rstep_mctxt_reflI)
    next
      assume "vars_term t \<inter> V = {} \<and> (s,t) \<in> rrstep (set R)"      
      then obtain l r \<sigma> where id: "s = l \<cdot> \<sigma>" "t = r \<cdot> \<sigma>" and 
        lr: "(l,r) \<in> set R" and
        vars: "vars_term t \<inter> V = {}" 
        by (metis rrstepE)      
      thus ?thesis by (intro exI[of _ MHole] exI[of _ "[Par_Info s t (l,r)]"], auto intro: par_rstep_mctxt_MHoleI)
    qed
  } note Prop = this  
  {
    assume "is_par_rstep_var_restr s t" 
    hence "\<exists> C infos. (s, t) \<in> par_rstep_mctxt (set R) C infos \<and> vars_below_hole t C \<inter> V = {}"
    proof (induct rule: is_par_rstep_var_restr.induct[])
      case "2_1" 
      thus ?case by (intro Prop, auto)
    next
      case "2_2" 
      thus ?case by (intro Prop, auto)
    next
      case (1 f ss g ts)
      show ?case
      proof (cases "?Prop (Fun f ss) (Fun g ts)")
        case True
        thus ?thesis using Prop by auto
      next
        case False
        with 1 have args: "f = g" "length ss = length ts" "list_all2 is_par_rstep_var_restr ss ts"
          by (auto split: if_splits)
        let ?P = "\<lambda> i C infos. (ss ! i, ts ! i) \<in> par_rstep_mctxt (set R) C infos \<and> vars_below_hole (ts ! i) C \<subseteq> (UNIV - V)" 
        { fix i
          assume i:"i < length ss"
          then have si:"ss ! i \<in> set ss" by auto
          from i args(2) have ti:"ts ! i \<in> set ts" by auto
          from args(3) have iprv:"is_par_rstep_var_restr (ss ! i) (ts ! i)" using i list_all2_nthD by blast
          with 1(1)[of "ss!i" "ts!i"] have pp:"\<exists>C infos. ?P i C infos" 
            using local.args(1) local.args(2) using si ti by blast
        } 
        hence "\<forall> i. \<exists> C infos. i < length ss \<longrightarrow> ?P i C infos" by blast
        from choice[OF this] obtain C where "\<forall> i. \<exists> infos. i < length ss \<longrightarrow> ?P i (C i) infos" by blast
        from choice[OF this] obtain infos where IH: "\<And> i. i < length ss \<Longrightarrow> ?P i (C i) (infos i)" by blast
        let ?C = "MFun f (map C [0..<length ss])" 
        let ?infos = "concat (map infos [0..<length ss])" 
        show ?thesis 
        proof (intro exI[of _ ?C] exI[of _ ?infos] conjI)
          show "vars_below_hole (Fun g ts) ?C \<inter> V = {}" using IH args(2) unfolding args(1)
            by (subst vars_below_hole_Fun; force)
          show "(Fun f ss, Fun g ts) \<in> par_rstep_mctxt (set R) ?C ?infos" unfolding args(1) using args(2) IH
            by (intro par_rstep_mctxt_funI, auto)
        qed
      qed
    qed
    thus " (s, t) \<in> par_rstep_var_restr (set R) V" unfolding par_rstep_var_restr_def by auto
  }
  {
    assume "(s, t) \<in> par_rstep_var_restr (set R) V" 
    from this[unfolded par_rstep_var_restr_def] obtain C infos where
       st: "(s, t) \<in> par_rstep_mctxt (set R) C infos" and vars: "vars_below_hole t C \<inter> V = {}" by auto
    thus "is_par_rstep_var_restr s t" 
    proof (induct C arbitrary: s t infos)
      case (MVar x)
      from par_rstep_mctxt_MVarE[OF MVar(1)]
      have "s = Var x" "t = Var x" by auto
      thus ?case by simp
    next
      case MHole
      from par_rstep_mctxt_MHoleE[OF MHole(1)]
      have "(s,t) \<in> rrstep (set R)" by auto  
      then show ?case using MHole(2) by (cases s; cases t; auto)
    next
      case (MFun f Cs)
      from par_rstep_mctxt_MFunD[OF MFun(2)]
      obtain ss ts Infos
        where s: "s = Fun f ss" and
          t: "t = Fun f ts" and 
          len: "length ss = length Cs" 
            "length ts = length Cs"
            "length Infos = length Cs" and
          infos: "infos = concat Infos" and
          steps: "\<And> i. i < length Cs \<Longrightarrow>(ss ! i, ts ! i) \<in> par_rstep_mctxt (set R) (Cs ! i) (Infos ! i)"
        by auto
      show ?case unfolding s t is_par_rstep_var_restr.simps
      proof (intro disjI2 conjI refl list_all2_all_nthI, unfold len)
        fix i
        assume i: "i < length Cs" 
        hence mem: "Cs ! i \<in> set Cs" by auto
        show "is_par_rstep_var_restr (ss ! i) (ts ! i)" 
        proof (rule MFun(1)[OF mem steps[OF i]])
          have "vars_below_hole (ts ! i) (Cs ! i) \<subseteq> vars_below_hole t (MFun f Cs)" 
            unfolding t using i len 
            by (subst vars_below_hole_Fun, auto)
          with MFun(3) show "vars_below_hole (ts ! i) (Cs ! i) \<inter> V = {}" by auto
        qed
      qed auto
    qed
  }
qed
end

lemma par_rstep_var_restr_code[code_unfold]: 
  "(s, t) \<in> par_rstep_var_restr (set R) V \<longleftrightarrow> is_par_rstep_var_restr R V s t" 
  by simp


subsection \<open>Implementation of Parallel Rewriting\<close>

fun is_par_rstep :: "('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
where
  "is_par_rstep R (Fun f ss) (Fun g ts) =
    (Fun f ss = Fun g ts \<or> (Fun f ss, Fun g ts) \<in> rrstep (set R) \<or>
    (f = g \<and> length ss = length ts \<and> list_all2 (is_par_rstep R) ss ts))"
| "is_par_rstep R s t = (s = t \<or> (s,t) \<in> rrstep (set R))"

lemma is_par_rstep[simp]:
  "is_par_rstep R s t \<longleftrightarrow> (s, t) \<in> par_rstep (set R)"
proof -
  have "is_par_rstep R s t = is_par_rstep_var_restr R {} s t"
    by (induct R s t rule: is_par_rstep.induct, auto simp del: is_par_rstep_var_restr simp: list_all2_conv_all_nth)
  also have "\<dots> \<longleftrightarrow> (s, t) \<in> par_rstep_var_restr (set R) {}" by simp
  also have "\<dots> \<longleftrightarrow> (s, t) \<in> par_rstep (set R)" 
    unfolding par_rstep_var_restr_def par_rstep_par_rstep_mctxt_conv by auto
  finally show ?thesis .
qed

lemma par_rstep_code[code_unfold]: "(s, t) \<in> par_rstep (set R) \<longleftrightarrow> is_par_rstep R s t" by simp


subsection \<open>Generate all root-rewrites (for well-formed TRS)\<close>

fun root_rewrite :: "('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term list"
where
  "root_rewrite R s = concat (map (\<lambda> (l, r).
    (case match s l of
      None \<Rightarrow> []
    | Some \<sigma> \<Rightarrow> [(r \<cdot> \<sigma>)])) R)"

lemma root_rewrite_sound:
  assumes "t \<in> set (root_rewrite R s)"
  shows "(s, t) \<in> rrstep (set R)"
proof -
  from assms
  have "\<exists> l r.  (l,r) \<in> set R \<and> t \<in> set (case match s l of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [r \<cdot> \<sigma>])"
    by auto
  from this obtain l r where one:
    "(l,r) \<in> set R \<and> t \<in> set (case match s l of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [r \<cdot> \<sigma>])"
    by auto
  from this obtain \<sigma> where two: "match s l = Some \<sigma> \<and> t \<in> {r \<cdot> \<sigma>}" by (cases "match s l", auto)
  then have match: "l \<cdot> \<sigma> = s" using match_sound by auto
  with one match one two have "(s,t) \<in> rstep_r_p_s (set R) (l,r) [] \<sigma>" unfolding rstep_r_p_s_def by (simp add: Let_def ctxt_supt_id)
  then show "(s,t) \<in> rrstep (set R)" unfolding rstep_iff_rstep_r_p_s rrstep_def by blast
qed

subsection \<open>Generate all parallel rewrites (for well formed TRS)\<close>
fun parallel_rewrite :: "('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term list"
where
  "parallel_rewrite R (Var x) = [Var x]"
| "parallel_rewrite R (Fun f ss) = remdups
     (root_rewrite R (Fun f ss) @ map (\<lambda>ss. Fun f ss) (product_lists (map (parallel_rewrite R) ss)))"

lemma parallel_rewrite_par_step:
  assumes "t \<in> set (parallel_rewrite R s)"
  shows "(s, t) \<in> par_rstep (set R)"
using assms
proof (induct s arbitrary: t)
  case (Fun f ss)
  then consider (root) "t \<in> set (root_rewrite R (Fun f ss))"
    | (args) "t \<in> set (map (\<lambda>ss. Fun f ss) (product_lists (map (parallel_rewrite R) ss)))"
    by force
  then show ?case
  proof (cases)
    case root
    from root_rewrite_sound[OF this] obtain l r \<sigma> where "(l, r) \<in> set R"
      and "l \<cdot> \<sigma> = Fun f ss" and "r \<cdot> \<sigma> = t"
      unfolding rrstep_def rstep_r_p_s_def by auto
    then show ?thesis using par_rstep.intros(1) by metis
  next
    case args
    then obtain ts where t:"t = Fun f ts" and ts:"ts \<in> set (product_lists (map (parallel_rewrite R) ss))"
      by auto
    then have len:"length ss = length ts" using in_set_product_lists_length by force
    { fix i
      assume i:"i < length ts"
      have "ts ! i \<in> set (parallel_rewrite R (ss ! i))"
        using ts[unfolded product_lists_set[of "_ ss"]]
        by (auto simp: list_all2_map2[of "(\<lambda>x ys. x \<in> set ys)"] intro: list_all2_nthD[OF _ i])
      with Fun(1) len i have "(ss ! i, ts ! i) \<in> par_rstep (set R)" by auto
    }
    from par_rstep.intros(2)[OF this len] show ?thesis using t by auto
  qed
qed auto

(*Takes a list of rewrite rules and two terms and checks for each rule whether the first term matches
the lhs and the second term the rhs of the rule. For each rule where this is the case,
returns the lists of terms corresponding to the matching substitutions. 
Terms that correspond to variables that only appear in the lhs of the rule (i.e. terms that are being
erased) don't play a role in the remainder of the computation for multisteps hence they can be ignored.*) 
fun root_steps_substs :: "('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> (('f, 'v) term list \<times> ('f, 'v) term list) list"
where
  "root_steps_substs R s t = concat (map (\<lambda> (l, r).
    (case match s l of
      None \<Rightarrow> []
    | Some \<sigma> \<Rightarrow> (case match t r of 
        None \<Rightarrow> []
      | Some \<tau> \<Rightarrow> [(map \<sigma> (var r), map \<tau> (var r))]))) 
  R)"

lemma root_steps_substs_exists:
  assumes "(ss, ts) \<in> set (root_steps_substs R s t)"
  shows "\<exists> l r \<sigma> \<tau>. (l, r) \<in> set R \<and> l \<cdot> \<sigma> = s \<and> r \<cdot> \<tau> = t \<and> (ss, ts) = (map \<sigma> (var r), map \<tau> (var r))"
proof-
from assms obtain l r where lr:"(l,r) \<in> set R" "(ss, ts) \<in> set (case match s l of
      None \<Rightarrow> []
    | Some \<sigma> \<Rightarrow> (case match t r of 
        None \<Rightarrow> []
      | Some \<tau> \<Rightarrow> [(map \<sigma> (var r), map \<tau> (var r))]))"
  by auto
  then obtain \<sigma> where \<sigma>:"match s l = Some \<sigma>"
    by fastforce 
  with lr obtain \<tau> where \<tau>:"match t r = Some \<tau>" 
    by fastforce
  from lr \<sigma> \<tau> have "(ss, ts) = (map \<sigma> (var r), map \<tau> (var r))"
    by simp 
  with lr(1) \<sigma> \<tau> show ?thesis
    using match_matches by blast
qed

lemma size_match_subst_Fun: (*used for termination_simps*)
  assumes "is_Fun l" and "x \<in> vars_term l"
  and match:"match s l = Some \<tau>"
shows "size (\<tau> x) < size s"
proof-
  from assms(1) obtain f ts where l:"l = Fun f ts"
    by blast 
  from match have *:"l \<cdot> \<tau> = s"
    by (simp add: match_matches) 
  then obtain ss where s:"s = Fun f ss"
    unfolding l by force
  from assms(2) obtain i where i:"i < length ts" and x:"x \<in> vars_term (ts!i)" 
    unfolding l by (metis term.sel(4) var_imp_var_of_arg) 
  from * have le:"length ts = length ss" 
    unfolding s l by auto 
  moreover from * i l s have "ts!i \<cdot> \<tau> = ss!i" 
    by fastforce
  then have "size (\<tau> x) \<le> size (ss!i)"
    using vars_term_size x by metis
  with i show ?thesis unfolding s term.size le
    by (metis add.commute add_0 add_Suc in_set_conv_nth less_Suc_eq_le size_list_estimation')
qed

fun root_subst_with_rhs :: "('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> (('f, 'v) term \<times> ('f, 'v) term list) list"
where
  "root_subst_with_rhs R s = concat (map (\<lambda> (l, r).
    (case match s l of
      None \<Rightarrow> []
    | Some \<sigma> \<Rightarrow> [(r, map \<sigma> (var r))]))
  R)"

lemma root_steps_subst_rhs_exists:
  assumes "(r, ss) \<in> set (root_subst_with_rhs R s)"
  shows "\<exists> l \<sigma>. (l, r) \<in> set R \<and> l \<cdot> \<sigma> = s \<and> ss = map \<sigma> (var r)"
proof-
from assms obtain l where lr:"(l,r) \<in> set R" "(r, ss) \<in> set (case match s l of
      None \<Rightarrow> []
    | Some \<sigma> \<Rightarrow> [(r, map \<sigma> (var r))])"
  by auto
  then obtain \<sigma> where \<sigma>:"match s l = Some \<sigma>"
    by fastforce
  with lr show ?thesis
    using match_matches by force
qed

context
  fixes R :: "('f, 'v) rules"
  assumes "wf_trs (set R)"
begin

private lemma *: "list_all (\<lambda>(l, r). is_Fun l \<and> (vars_term r \<subseteq> vars_term l)) R" 
  using \<open>wf_trs (set R)\<close> unfolding wf_trs_def by (auto simp: list_all_iff)

lemma varcond:
  "\<And>l r. (l, r) \<in> set R \<Longrightarrow> is_Fun l \<and> vars_term r \<subseteq> vars_term l"
  using * Ball_set_list_all case_prodD by (metis (no_types, lifting)) 

lemma [termination_simp]:
  assumes "(l, r) \<in> set R" 
    and "Some \<sigma> = match (Fun g ts) l" 
    and "x \<in> vars_term r"
  shows "size (\<sigma> x) < Suc (size_list size ts)"
  using assms size_match_subst_Fun varcond
  by (metis (no_types, lifting) add.right_neutral add_Suc_right subsetD term.size(4))

lemma [termination_simp]:
  assumes "(l, r) \<in> set R"
    and "Some \<tau> = match s l" 
    and "x \<in> vars_term r" 
  shows "size (\<tau> x) < size s"
  using assms size_match_subst_Fun varcond by (metis subset_iff)

lemma [termination_simp]:
  assumes "(l, r) \<in> set R"
    and "Some \<tau> = match (Var v) l" 
  shows "size (\<tau> x) = 0"
proof-
  from assms(1) varcond obtain f ss where "l = Fun f ss"
    by blast 
  with assms(2) have False
    by (metis match_matches subst_apply_eq_Var term.distinct(1)) 
  then show ?thesis 
    by simp
qed

text\<open>Given two terms s and t determine whether there exists a multi-step from s to t in the given TRS.\<close>
(*Only terminating if variable left-hand sides are excluded.*)
fun is_mstep_main :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
  where
  "is_mstep_main (Fun f ss) (Fun g ts) =
    (Fun f ss = Fun g ts \<or> 
    list_ex (\<lambda> (ss, ts). list_all2 is_mstep_main ss ts) (root_steps_substs R (Fun f ss) (Fun g ts)) \<or> 
    (if f = g \<and> length ss = length ts then list_all2 is_mstep_main ss ts else False))" 
| "is_mstep_main s t = (s = t \<or> list_ex (\<lambda> (ss, ts). list_all2 is_mstep_main ss ts) (root_steps_substs R s t))"

text\<open>Show that all multisteps are covered by the definition above.\<close>
lemma mstep_is_mstep_main:
  assumes "(s, t) \<in> mstep (set R)"
  shows "is_mstep_main s t"
  using assms proof(induct)
  case (args f n ss ts)
  then have "list_all2 is_mstep_main ss ts"
    by (simp add: list_all2_all_nthI) 
  with args show ?case
    by simp 
next
  case (rule l r \<sigma> \<tau>)
  from rule(1) obtain i where i:"i < length R" "R!i = (l, r)"
    by (metis in_set_idx) 
  obtain \<sigma>' where sigma:"match (l \<cdot> \<sigma>) l = Some \<sigma>'" "(\<forall>x\<in>vars_term l. \<sigma> x = \<sigma>' x)" 
    by (meson match_complete')
  obtain \<tau>' where tau:"match (r \<cdot> \<tau>) r = Some \<tau>'" "(\<forall>x\<in>vars_term r. \<tau> x = \<tau>' x)" 
    by (meson match_complete')
  let ?matches="(map (\<lambda>(l', r'). case match (l \<cdot> \<sigma>) l' of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> case match (r \<cdot> \<tau>) r' of None \<Rightarrow> [] | Some \<tau> \<Rightarrow> [(map \<sigma> (var r'), map \<tau> (var r'))]) R)"
  have "i < length ?matches"
    using i(1) by auto 
  moreover have "(map \<sigma>' (var r), map \<tau>' (var r)) \<in> set (?matches ! i)"
    using sigma(1) tau(1) i by simp
  ultimately have "(map \<sigma>' (var r), map \<tau>' (var r)) \<in> set (root_steps_substs R (l\<cdot>\<sigma>) (r\<cdot>\<tau>))"
    unfolding root_steps_substs.simps by (metis (no_types, lifting) concat_nth concat_nth_length in_set_conv_nth) 
  then obtain j where j:"j < length (root_steps_substs R (l\<cdot>\<sigma>) (r\<cdot>\<tau>))" "root_steps_substs R (l\<cdot>\<sigma>) (r\<cdot>\<tau>) ! j = (map \<sigma>' (var r), map \<tau>' (var r))"
    by (metis in_set_idx)
  have "(\<lambda> (ss, ts). list_all2 is_mstep_main ss ts) ((root_steps_substs R (l\<cdot>\<sigma>) (r\<cdot>\<tau>))!j)"
  proof-
    {fix i assume "i < length (var r)"
      then have *:"var r!i \<in> vars_term r"
        by (metis comp_apply nth_mem set_remdups set_rev set_vars_term_list) 
      then have "var r!i \<in> vars_term l" 
        using varcond rule.hyps(1) by blast 
      moreover then have "\<sigma>' (var r !i) = \<sigma> (var r ! i)" 
        using sigma(2) * by simp
      moreover have "\<tau>' (var r !i) = \<tau> (var r ! i)"
        using * tau(2) by simp
      ultimately have "is_mstep_main (\<sigma>' (var r !i)) (\<tau>' (var r!i))" 
        using rule(2) by force
    }
    then have "list_all2 is_mstep_main (map \<sigma>' (var r)) (map \<tau>' (var r))"
      by (simp add: list_all2_conv_all_nth) 
    then show ?thesis 
      unfolding j(2) by fastforce 
  qed
  then have *:"list_ex (\<lambda> (ss, ts). list_all2 is_mstep_main ss ts) (root_steps_substs R (l\<cdot>\<sigma>) (r\<cdot>\<tau>))"
    using j by (meson list_ex_length) 
  then show ?case proof(cases "l \<cdot> \<sigma>")
    case (Var x)
    then show ?thesis 
      using * is_mstep_main.simps(2) by presburger 
  next
    case (Fun f ts)
    note f=this
    show ?thesis proof(cases "r \<cdot> \<tau>")
      case (Var x)
      then show ?thesis 
        using * is_mstep_main.simps(3) by presburger 
    next
      case (Fun g ss)
      then show ?thesis 
        using f * is_mstep_main.simps(1) by presburger 
    qed
  qed 
qed simp

lemma is_mstep_main:
  assumes "is_mstep_main s t"
  shows "(s, t) \<in> mstep (set R)"
using assms proof (induct rule: is_mstep_main.induct)
  case (1 f ss g ts)
  then consider (eq) "Fun f ss = Fun g ts"
    | (root) "list_ex (\<lambda> (ss, ts). list_all2 is_mstep_main ss ts) (root_steps_substs R (Fun f ss) (Fun g ts))"
    | (args) "f = g" and "length ss = length ts" and "list_all2 is_mstep_main ss ts"
    by (auto split: if_splits)
  then show ?case proof(cases)
    case root
    then obtain i where "i < length (root_steps_substs R (Fun f ss) (Fun g ts))" "(\<lambda> (ss, ts). list_all2 is_mstep_main ss ts) ((root_steps_substs R (Fun f ss) (Fun g ts))!i)"
      using list_ex_length by blast
    then obtain ss' ts' where ss'ts':"(ss' , ts') \<in> set (root_steps_substs R (Fun f ss) (Fun g ts))" "list_all2 is_mstep_main ss' ts'"
      using nth_mem by fastforce 
    with root_steps_substs_exists obtain l r \<sigma> \<tau> where lr:"(l, r) \<in> set R" 
      and l:"l \<cdot> \<sigma> = Fun f ss" and r:"r \<cdot> \<tau> = Fun g ts" 
      and \<sigma>\<tau>:"(ss', ts') = (map \<sigma> (var r), map \<tau> (var r))"
      by blast
    let ?\<tau>="\<lambda>x. (if x \<in> vars_term r then \<tau> x else \<sigma> x)"
    from r have r':"r \<cdot> ?\<tau> = Fun g ts"
      by (smt (verit, del_insts) term_subst_eq) 
    { fix x assume "x \<in> vars_term l"
      then have "(\<sigma> x, ?\<tau> x) \<in> mstep (set R)" proof(cases "x \<in> vars_term r")
        case True
        then obtain i where i:"i < length (var r)" "var r ! i = x"
          by (metis in_set_idx set_vars_term_list vars_term_list_var)  
        then have i1:"i < length ss'" 
          using \<sigma>\<tau> by simp 
        from i have i2:"i < length ts'"
          using \<sigma>\<tau> by simp
        from ss'ts'(2) i1 i2 have "is_mstep_main (ss' ! i) (ts' ! i)"
          using list_all2_nthD by blast
        with 1(1)[OF ss'ts'(1)] i1 i2 have "(ss' ! i, ts' ! i) \<in> mstep (set R)"
          by auto
        then show ?thesis 
          using i \<sigma>\<tau> by auto 
      qed simp
    }
    then show ?thesis 
      using mstep.rule[OF lr] l r' by force 
  next
    case args
    { fix i
      assume i:"i < length ss"
      then have si:"ss ! i \<in> set ss" by auto
      from i args(2) have ti:"ts ! i \<in> set ts" by auto
      from args(3) have "is_mstep_main (ss ! i) (ts ! i)" using i list_all2_nthD by blast
      with 1(2)[of "ss ! i" "ts ! i"] args(1,2) si ti have "(ss ! i, ts ! i) \<in> mstep (set R)"
        by auto
    }
    then show ?thesis using args(1,2)
      by (simp add: mstep.args)
  qed simp
next
  case ("2_1" v t) 
  then consider (eq) "Var v = t"
    | (root) "list_ex (\<lambda> (ss, ts). list_all2 is_mstep_main ss ts) (root_steps_substs R (Var v) t)"
    by auto
  then show ?case proof(cases)
    case root
    then obtain i where "i < length (root_steps_substs R (Var v) t)" "(\<lambda> (ss, ts). list_all2 is_mstep_main ss ts) ((root_steps_substs R (Var v) t)!i)"
      unfolding is_mstep_main.simps using list_ex_length by blast
    then obtain ss' ts' where ss'ts':"(ss' , ts') \<in> set (root_steps_substs R (Var v) t)" "list_all2 is_mstep_main ss' ts'"
      using nth_mem by fastforce 
    with root_steps_substs_exists obtain l r \<sigma> \<tau> where lr:"(l, r) \<in> set R" 
      and l:"l \<cdot> \<sigma> = Var v" and r:"r \<cdot> \<tau> = t" 
      and \<sigma>\<tau>:"(ss', ts') = (map \<sigma> (var r), map \<tau> (var r))"
      by blast
    let ?\<tau>="\<lambda>x. (if x \<in> vars_term r then \<tau> x else \<sigma> x)"
    from r have r':"r \<cdot> ?\<tau> = t"
      by (smt (verit, del_insts) term_subst_eq) 
    { fix x assume "x \<in> vars_term l"
      then have "(\<sigma> x, ?\<tau> x) \<in> mstep (set R)" proof(cases "x \<in> vars_term r")
        case True
        then obtain i where i:"i < length (var r)" "var r ! i = x"
          by (metis in_set_idx set_vars_term_list vars_term_list_var)  
        then have i1:"i < length ss'" 
          using \<sigma>\<tau>  by simp 
        from i have i2:"i < length ts'"
          using \<sigma>\<tau> by simp
        from ss'ts'(2) i1 i2 have "is_mstep_main (ss' ! i) (ts' ! i)"
          using list_all2_nthD by blast
        with "2_1"(1)[OF ss'ts'(1)] i1 i2 have "(ss' ! i, ts' ! i) \<in> mstep (set R)"
          by auto
        then show ?thesis
          using i \<sigma>\<tau> by auto 
      qed simp
    }
    then show ?thesis 
      using mstep.rule[OF lr] l r' by force 
  qed simp
next
  case ("2_2" s v)
  then consider (eq) "s = Var v"
    | (root) "list_ex (\<lambda> (ss, ts). list_all2 is_mstep_main ss ts) (root_steps_substs R s (Var v))"
    by auto
  then show ?case proof(cases)
    case root
    then obtain i where "i < length (root_steps_substs R s (Var v))" "(\<lambda> (ss, ts). list_all2 is_mstep_main ss ts) ((root_steps_substs R s (Var v))!i)"
      unfolding is_mstep_main.simps using list_ex_length by blast
    then obtain ss' ts' where ss'ts':"(ss' , ts') \<in> set (root_steps_substs R s (Var v))" "list_all2 is_mstep_main ss' ts'"
      using nth_mem by fastforce 
    with root_steps_substs_exists obtain l r \<sigma> \<tau> where lr:"(l, r) \<in> set R" 
      and l:"l \<cdot> \<sigma> = s" and r:"r \<cdot> \<tau> = Var v" 
      and \<sigma>\<tau>:"(ss', ts') = (map \<sigma> (var r), map \<tau> (var r))"
      by blast
    let ?\<tau>="\<lambda>x. (if x \<in> vars_term r then \<tau> x else \<sigma> x)"
    from r have r':"r \<cdot> ?\<tau> = Var v"
      by (smt (verit, del_insts) term_subst_eq) 
    { fix x assume "x \<in> vars_term l"
      then have "(\<sigma> x, ?\<tau> x) \<in> mstep (set R)" proof(cases "x \<in> vars_term r")
        case True
        then obtain i where i:"i < length (var r)" "var r ! i = x"
          by (metis in_set_idx set_vars_term_list vars_term_list_var)  
        then have i1:"i < length ss'" 
          using \<sigma>\<tau>  by simp 
        from i have i2:"i < length ts'"
          using \<sigma>\<tau> by simp
        from ss'ts'(2) i1 i2 have "is_mstep_main (ss' ! i) (ts' ! i)"
          using list_all2_nthD by blast
        with "2_2"(1)[OF ss'ts'(1)] i1 i2 have "(ss' ! i, ts' ! i) \<in> mstep (set R)"
          by auto
        then show ?thesis
          using i \<sigma>\<tau> by auto 
      qed simp
    }
    then show ?thesis 
      using mstep.rule[OF lr] l r' by force 
  qed simp
qed

lemma mstep_iff_is_mstep_main: "(s,t) \<in> mstep (set R) \<longleftrightarrow> is_mstep_main s t" 
  using is_mstep_main[of s t] mstep_is_mstep_main[of s t] by auto
  

(*compute list of terms reachable in multi-step from a given term.*)
fun mstep_rewrite_main :: "('f, 'v) term \<Rightarrow> ('f, 'v) term list"
where
  "mstep_rewrite_main (Var x) = [Var x]"
| "mstep_rewrite_main (Fun f ss) = remdups (
     (concat (map (\<lambda>(r, ts). 
        (map (\<lambda>args. r \<cdot> (mk_subst Var (zip (var r) args))) (product_lists (map mstep_rewrite_main ts))))
      (root_subst_with_rhs R (Fun f ss))))
    @(map (\<lambda>ss. Fun f ss) (product_lists (map mstep_rewrite_main ss))))"

lemma mstep_rewrite_main_mstep:
  assumes "t \<in> set (mstep_rewrite_main s)"
  shows "(s, t) \<in> mstep (set R)"
using assms
proof (induct s arbitrary: t rule:subterm_induct)
  case (subterm s)
  then show ?case proof(cases s)
    case (Var x)
    with subterm(2) show ?thesis by simp
  next
    case (Fun f ss)
    with subterm consider (root) "t \<in> set (concat (map (\<lambda>(r,ts).(map (\<lambda>args. r \<cdot> (mk_subst Var (zip (var r) args))) 
          (product_lists (map mstep_rewrite_main ts)))) (root_subst_with_rhs R (Fun f ss))))"
      | (args) "t \<in> set (map (\<lambda>ss. Fun f ss) (product_lists (map mstep_rewrite_main ss)))"
      by force 
    then show ?thesis
    proof (cases)
      case root
      then obtain r ts where rhs_subst:"(r,ts) \<in> set (root_subst_with_rhs R (Fun f ss))" 
                             "t \<in> set (map (\<lambda>args. r \<cdot> (mk_subst Var (zip (var r) args))) (product_lists (map mstep_rewrite_main ts)))"
        by force
      from root_steps_subst_rhs_exists[OF rhs_subst(1)] obtain l \<sigma> where lr:"(l, r) \<in> set R"
        and sigma:"l \<cdot> \<sigma> = Fun f ss" "ts = map \<sigma> (var r)" by auto
      from rhs_subst(2) obtain args where args:"t = r \<cdot> (mk_subst Var (zip (var r) args))"
                                      "args \<in> set (product_lists (map mstep_rewrite_main ts))"
        by auto    
      then have len:"length args = length ts"
        using in_set_product_lists_length by fastforce 
      then have len':"length args = length (var r)"
        by (simp add: sigma(2)) 
      let ?\<tau>="\<lambda>x. if x \<in> vars_term r then (mk_subst Var (zip (var r) args)) x else \<sigma> x"
      from args(1) have t:"t = r \<cdot> ?\<tau>"
        by (simp add: term_subst_eq_conv) 
      { fix x
        assume x:"x \<in> vars_term l"
        have "(\<sigma> x, ?\<tau> x) \<in> mstep (set R)" proof(cases "x \<in> vars_term r")
          case True
          then obtain i where i:"i < length (var r)" "x = var r ! i"
            by (metis in_set_idx set_vars_term_list vars_term_list_var)
          with True len' have tau_x:"?\<tau> x = args!i"
            by (simp add: mk_subst_distinct)
          from i sigma(2) have sigma_x:"\<sigma> x = ts!i"
            by simp 
          have "\<sigma> x \<lhd> Fun f ss"
            by (metis is_VarI lr sigma(1) subst_image_subterm term.set_cases(2) varcond x) 
          with sigma_x have "ts!i \<lhd> Fun f ss" by simp
          moreover have "args!i \<in> set (mstep_rewrite_main (ts!i))" using args(2) i(1) len' len 
            unfolding product_lists_set list_all2_conv_all_nth by force
          ultimately show ?thesis using subterm(1) sigma_x tau_x unfolding Fun by presburger
        next
          case False 
          then show ?thesis by simp
        qed
      }
      then show ?thesis using mstep.intros(3)[OF lr] sigma(1) unfolding Fun t
        by fastforce
    next
      case args
      then obtain ts where t:"t = Fun f ts" and ts:"ts \<in> set (product_lists (map mstep_rewrite_main ss))"
        by auto
      then have len:"length ss = length ts" using in_set_product_lists_length by force
      { fix i
        assume i:"i < length ts"
        have "ts ! i \<in> set (mstep_rewrite_main (ss ! i))"
          using ts[unfolded product_lists_set[of "_ ss"]]
          by (auto simp: list_all2_map2[of "(\<lambda>x ys. x \<in> set ys)"] intro: list_all2_nthD[OF _ i])
        with subterm len i have "(ss ! i, ts ! i) \<in> mstep (set R)" 
          unfolding Fun by auto
      }
      with mstep.intros(2) len t Fun show ?thesis
        by metis
    qed
  qed
qed

end

text\<open>We need to be able to export code for @{const is_mstep_main} and @{const mstep_rewrite_main}, hence the following definitions
(adapted from template by Rene).\<close>
(*New type for well-formed TRSs*)
typedef ('f, 'v) wfTRS = "{R :: ('f, 'v) rules. wf_trs (set R)}"
  by (intro exI[of _ Nil], auto simp: wf_trs_def)

setup_lifting type_definition_wfTRS

lift_definition get_TRS :: "('f, 'v) wfTRS \<Rightarrow> ('f, 'v) rules" is "\<lambda> R. R" .

lemma is_wf_get_TRS: "wf_trs (set (get_TRS R'))" 
  by (transfer, auto)

definition "is_mstep_wf R = is_mstep_main (get_TRS R)" 
definition "mstep_rewrite_wf R = mstep_rewrite_main (get_TRS R)" 

lemmas is_mstep_wf_simps = is_mstep_main.simps[OF is_wf_get_TRS, folded is_mstep_wf_def]
declare is_mstep_wf_simps[code]

lemmas mstep_rewrite_wf_simps = mstep_rewrite_main.simps[OF is_wf_get_TRS, folded mstep_rewrite_wf_def]
declare mstep_rewrite_wf_simps[code]

(* one might use an implementation which does not require show-class *)
lift_definition (code_dt) get_wfTRS :: "('f :: showl, 'v :: showl) rules \<Rightarrow> ('f, 'v) wfTRS option" is
  "\<lambda> R. if isOK (check_wf_trs R) then Some R else None"
  by (force simp: wf_trs_def list.pred_set split: prod.splits)

definition err_wf where "err_wf = STR ''TRS is not well-formed''" 
(*should actually never be printed, since TRS is checked before even calling is_mstep_main*) 

definition "mstep_dummy_impl R s t = ((s,t) \<in> mstep (set R))" 
lemma mstep_dummy_impl[code]: "mstep_dummy_impl R = Code.abort (STR ''mstep_dummy'') (\<lambda> _. mstep_dummy_impl R)" 
  by simp

definition "is_mstep R = (case get_wfTRS R of Some R' \<Rightarrow> is_mstep_wf R' 
  | None \<Rightarrow> Code.abort err_wf (\<lambda> _. mstep_dummy_impl R))" 

lemma is_mstep[simp]: "is_mstep R = (\<lambda> s t. (s,t) \<in> mstep (set R))" 
proof (cases "wf_trs (set R)")
  case False
  hence "get_wfTRS R = None" by transfer auto
  thus ?thesis by (auto simp: mstep_dummy_impl_def[abs_def] is_mstep_def)
next
  case True  
  hence "\<exists> R'. get_wfTRS R = Some R' \<and> get_TRS R' = R"
    by (transfer, auto)
  thus ?thesis unfolding is_mstep_def is_mstep_wf_def using mstep_iff_is_mstep_main[OF True] 
    by auto
qed


lift_definition (code_dt) get_wfTRS_sub :: "('f :: showl, 'v :: showl) rules \<Rightarrow> ('f, 'v) wfTRS" is
  "\<lambda> R. if isOK (check_wf_trs R) then R else Code.abort err_wf (\<lambda> _. [])"
  by (auto simp: wf_trs_def)

definition "mstep_rewrite R = mstep_rewrite_wf (get_wfTRS_sub R)" 

lemma mstep_rewrite_mstep:
  assumes "t \<in> set (mstep_rewrite R s)"
  shows "(s, t) \<in> mstep (set R)"
proof -
  define R' where "R' = get_wfTRS_sub R" 
  have wf: "wf_trs (set (get_TRS R'))" 
    by (transfer, auto)
  have sub: "set (get_TRS R') \<subseteq> set R" unfolding R'_def by (transfer, auto)
  from mstep_rewrite_main_mstep[OF wf, folded mstep_rewrite_wf_def, OF assms(1)[unfolded mstep_rewrite_def, folded R'_def]]
  have "(s, t) \<in> mstep (set (get_TRS R'))" .
  with mstep_mono[OF sub] show ?thesis  by auto
qed


end
